// Declare global constants
var gv_canvas = document.getElementById("gameCanvas");
var gv_context = gv_canvas.getContext("2d");

var gv_requestAnimationFrame;
var gv_isGameInProgress = false;
var gv_isGameEnding = false;

var gv_images = {};
var gv_numImagesLoaded = 0;
var gv_numImagesRequested = 4;

var gv_ship = null;

var gv_enemies = [];
var gv_numEnemies = 10;
var gv_enemySpeed = 1.0;

var gv_bullets = [];
var gv_numBullets = 10;
var gv_bulletSeparation = 5;  // pixels
var gv_bulletSpeed = 5.0;
var gv_bulletCreationTimer = null;
var gv_isFiringBullets = false;

// variables for audio
var gv_hurtSound = gameAudio.createSound("hurt");
var gv_laserSound = gameAudio.createSound("laser");
var gv_playerHurtSound = gameAudio.createSound("playerHurt");
var gv_isSoundOn = false;

var gv_score = 0;

var gv_playerDifficulties = [
    { difficulty: "easy", numBullets: 18, numEnemies: 50, enemySpeed: 1.0 },
    { difficulty: "medium", numBullets: 12, numEnemies: 100, enemySpeed: 2.0 },
    { difficulty: "hard", numBullets: 6, numEnemies: 150, enemySpeed: 3.0 }
];

function setDifficulty(level) {
    var ii;
    for (ii = 0; ii < gv_playerDifficulties.length; ii++) {
        if (gv_playerDifficulties[ii].difficulty === level) {
            gv_numBullets = gv_playerDifficulties[ii].numBullets;
            gv_numEnemies = gv_playerDifficulties[ii].numEnemies;
            gv_enemySpeed = gv_playerDifficulties[ii].enemySpeed;
            break;
        }
    }
}

/**
 * Log information to the screen.
 *
 * @param text The text to display in the log area.
 */
function log(text) {
    var logArea = document.getElementById("log");
    logArea.innerHTML = text;
}

function reenableControls() {
    var startButton = document.getElementById("start_button");
    var difficultySelector = document.getElementById("difficulty");

    difficultySelector.disabled = false;
    startButton.style.visibility = "visible";
}

function setGameOver() {
    window.cancelAnimationFrame(gv_requestAnimationFrame);
    log("<span class='gameover'>Game Over<\/span>");
    gv_isGameInProgress = false;
    clearScreen();
    reenableControls();
}

function setGameWon() {
    window.cancelAnimationFrame(gv_requestAnimationFrame);
    log("<span class='gameover'>Congratulations, You Won!<\/span>");
    gv_isGameInProgress = false;
    clearScreen();
    reenableControls();
}

/**
 * Process key down events.
 * This is how the ship's position is adjusted, how bullets are fired, and how the game is ended.
 *
 * @param event The key down event.
 */
function onKeyDown(event) {
    if (event.keyCode === 37) {  // left arrow
        gv_ship.isMovingLeft = true;
        event.preventDefault();
    } else if (event.keyCode === 39) {  // right arrow
        gv_ship.isMovingRight = true;
        event.preventDefault();
    } else if (event.keyCode === 32) {  // space key
        gv_isFiringBullets = true;
        event.preventDefault();
    } else if (event.keyCode === 27) {  // escape key
        setGameOver();  // End the game
    }
}

/**
 * Process key up events.
 * This is how the ship's position is adjusted, how bullets are fired, and how the game is ended.
 *
 * @param event The key up event.
 */
function onKeyUp(event) {
    if (event.keyCode === 37) {  // left arrow
        event.preventDefault();
        gv_ship.isMovingLeft = false;
    } else if (event.keyCode === 39) {  // right arrow
        event.preventDefault();
        gv_ship.isMovingRight = false;
    } else if (event.keyCode === 32) {  // space key
        event.preventDefault();
        gv_isFiringBullets = false;
    } else if (event.keyCode === 27) {  // escape key
        // End the game
    }
}

function fireBulletCallback() {
    if (gv_bulletCreationTimer != null) {
        gv_bulletCreationTimer = null;
    }
}

function fireBullet() {
    var ii;

    if (gv_isFiringBullets) {
        if (null === gv_bulletCreationTimer) {
            gv_bulletCreationTimer = setTimeout(fireBulletCallback, 150);
            for (ii = 0; ii < gv_bullets.length; ii++) {
                if (! gv_bullets[ii].isAlive) {
                    gv_bullets[ii].x = gv_ship.x;
                    gv_bullets[ii].y = gv_ship.y - gv_ship.h2;
                    gv_bullets[ii].isAlive = true;
                    if (gv_isSoundOn) {
                        gv_laserSound.play();
                    }
                    break;
                }
            }
        }
    } else if (gv_bulletCreationTimer != null) {
        // The fire button is not pressed - clear the timer.
        clearTimeout(gv_bulletCreationTimer);
        gv_bulletCreationTimer = null;
    }
}

function updateBullets() {
    var ii;

    if (gv_isFiringBullets && (null === gv_bulletCreationTimer)) {
        fireBullet();
    }

    for (ii = 0; ii < gv_bullets.length; ii++) {
        if (gv_bullets[ii].isAlive) {
            gv_bullets[ii].update();
        }
    }
}

function spawnEnemy() {
    var ii;

    for (ii = 0; ii < gv_enemies.length; ii++) {
        if (! gv_enemies[ii].isSpawned) {
            gv_enemies[ii].getRandomStart();
            gv_enemies[ii].isAlive = true;
            gv_enemies[ii].isSpawned = true;
            setTimeout(spawnEnemy, utils.getRandomInt(50, 500));
            break;
        }
    }
}

function updateEnemies() {
    var ii;

    for (ii = 0; ii < gv_enemies.length; ii++) {
        if (gv_enemies[ii].isAlive) {
            gv_enemies[ii].update();
        }
    }
}

function Actor(image, direction) {
    this.image = image;
    this.direction = direction;
    this.w2 = image.width / 2;  // pre-compute for speed.
    this.h2 = image.height / 2;  // pre-compute for speed.
    this.rSquared = (this.w2 * this.w2) + (this.h2 * this.h2);  // radius squared.
    this.x = this.w2;
    this.y = this.h2;
    this.speed = 0;
    this.angle = 0;
    this.xFactor = 0;
    this.yFactor = 0;
    this.deltaAccel = 0.5;
    this.friction = 0.90;
    this.isMovingLeft = false;
    this.isMovingRight = false;
    this.isAlive = true;
    this.isExploding = false;
    this.globalAlpha = 1.0;

    this.computeAngle = function () {
        if (this.direction === "right") {
            this.angle = 0;
        } else if (this.direction === "left") {
            this.angle = Math.PI;
        } else if (this.direction === "up") {
            this.angle = -Math.PI / 2;
        } else {  // "down"
            this.angle = Math.PI / 2;
        }
        this.xFactor = Math.cos(this.angle);  // pre-compute for speed.
        this.yFactor = Math.sin(this.angle);  // pre-compute for speed.
    };
    this.computeAngle();

    /**
     * Draw the object at its current location and rotation.
     */
    this.draw = function() {
        if (this.isExploding) {
            gv_context.save();
            gv_context.globalAlpha = this.globalAlpha;
            gv_context.drawImage(gv_images.explosion, this.x - (gv_images.explosion.width / 2), this.y - (gv_images.explosion.height / 2));
            gv_context.restore();
            this.globalAlpha -= 0.01;
            if (this.globalAlpha <= 0) {
                this.isAlive = false;
            }
        } else {
            gv_context.drawImage(this.image, this.x - this.w2, this.y - this.h2);
        }
    };

    this.edgeDetect = function() {
    };

    this.updateSpeed = function() {
    };

    /**
     * Update the actor's speed, rotation, and position.
     */
    this.update = function() {
        if (! this.isAlive) {
            return;
        }

        this.updateSpeed();

        this.speed *= this.friction;

        if ((this.speed > -0.005) && (this.speed < 0.005)) {
            this.speed = 0.0;  // Ensure braking to zero.
        }

        if (! gv_isGameEnding) {
            this.x += this.speed * this.xFactor;
            this.y += this.speed * this.yFactor;
            this.edgeDetect();
        }

        this.draw();
    };

    this.hitTest = function(object2) {
        var dx = object2.x - this.x;  // X-distance between objects
        var dy = object2.y - this.y;  // Y-distance between objects
        var rSquared = (dx * dx) + (dy * dy);  // Square of distance between objects
        // Is square of distance between objects <= sum of object radii squared?  (Yes == hit).
        return (rSquared <= (this.rSquared + object2.rSquared));
    };

    this.explode = function() {
        this.isExploding = true;
        this.friction = 0;  // stop all movement.
    };

}  // end: Actor

function hitTestShipByEnemies() {
    var ii;

    for (ii = 0; ii < gv_enemies.length; ii++) {
        if (gv_enemies[ii].isAlive) {
            if (gv_ship.hitTest(gv_enemies[ii])) {
                gv_isGameEnding = true;
                gv_ship.explode();
                if (gv_isSoundOn) {
                    gv_playerHurtSound.play();
                }
            }
        }
    }
}

function hitTestEnemiesByBullets() {
    var ii;
    var jj;

    for (ii = 0; ii < gv_bullets.length; ii++) {
        if (gv_bullets[ii].isAlive) {
            for (jj = 0; jj < gv_enemies.length; jj++) {
                if (gv_enemies[jj].isExploding) {
                    continue;
                }
                if (gv_enemies[jj].isAlive) {
                    if (gv_bullets[ii].hitTest(gv_enemies[jj])) {
                        gv_bullets[ii].isAlive = false;
                        gv_enemies[jj].explode();
                        if (gv_isSoundOn) {
                            gv_hurtSound.play();
                        }
                        gv_score += 10;
                        break;
                    }
                }
            }
        }
    }
}

function displayScore() {
    // var text = "Score: " + gv_score + " dead: " + getNumEnemiesDead() + " active: " + getNumEnemiesActive() + " total: " + gv_numEnemies;
    var text = "Score: " + gv_score;
    var textSize;

    gv_context.save();
    gv_context.strokeStyle = "#FFFFFF";
    gv_context.font = "30px Arial";
    textSize = gv_context.measureText(text);
    gv_context.strokeText(text, (gv_canvas.width - textSize.width) / 2, gv_canvas.height - 20);
    gv_context.restore();
}

function getNumEnemiesActive() {
    var ii;
    var numEnemiesActive = 0;

    for (ii = 0; ii < gv_enemies.length; ii++) {
        if ((! gv_enemies[ii].isExploding) && (gv_enemies[ii].isAlive)) {
            numEnemiesActive++;
        }
    }
    return numEnemiesActive;
}

function getNumEnemiesDead() {
    var ii;
    var numEnemiesDead = 0;

    for (ii = 0; ii < gv_enemies.length; ii++) {
        if (gv_enemies[ii].isExploding && (! gv_enemies[ii].isAlive)) {
            numEnemiesDead++;
        }
    }
    return numEnemiesDead;
}

function checkIfGameWon() {
    var ii;
    var numEnemiesDead = 0;

    for (ii = 0; ii < gv_enemies.length; ii++) {
        if (gv_enemies[ii].isExploding && (! gv_enemies[ii].isAlive)) {
            numEnemiesDead++;
        }
    }
    if ((gv_enemies.length > 0) && (numEnemiesDead >= gv_numEnemies)) {
        setGameWon();
    }
}

function clearScreen() {
    // Draw the background
    gv_context.fillStyle = "#111111";  // "#ECECEC";
    gv_context.fillRect(0, 0, gv_canvas.width, gv_canvas.height);
    gv_context.strokeStyle = "#7F7F7F";
    gv_context.strokeRect(0, 0, gv_canvas.width, gv_canvas.height);
}

function doUpdateGame() {
    clearScreen();

    gv_ship.update();

    updateBullets();
    updateEnemies();

    hitTestShipByEnemies();

    if (! gv_ship.isAlive) {
        setGameOver();
    }

    hitTestEnemiesByBullets();

    displayScore();
    checkIfGameWon();
}

/**
 * This function is called once every update of the animation frame.
 * It draws the background, updates the object positions, draws the objects,
 * and displays the score.
 */
function updateGame() {
    if (! gv_isGameInProgress) {
        return;
    }
    doUpdateGame();

}

function initShip() {
    gv_ship.x = gv_canvas.width / 2;
    gv_ship.y = gv_canvas.height - 80;
    gv_ship.friction = 0.90;
    gv_ship.isAlive = true;
    gv_ship.isExploding = false;
}

function initBullets() {
    var bullet;
    var ii;

    gv_bullets = [];

    for (ii = 0; ii < gv_numBullets; ii++) {
        bullet = new Actor(gv_images.bullet, "up");
        bullet.speed = gv_bulletSpeed;
        bullet.friction = 1.0;
        bullet.isAlive = false;
        bullet.edgeDetect = function() {
            if (this.y <= 0) {
                this.isAlive = false;
            }
        };
        bullet.hitTest = function(object2) {
            var dy;
            var dySquared;
            var dxLeft;
            var dxRight;
            var rSquaredLeft;
            var rSquaredRight;

            dy = object2.y - this.y;  // Y-distance between objects
            dySquared = dy * dy;
            dxLeft = object2.x - this.x - gv_bulletSeparation;  // X-distance between objects (left bullet)
            dxRight = object2.x - this.x + gv_bulletSeparation;  // X-distance between objects (right bullet)
            rSquaredLeft = (dxLeft * dxLeft) + dySquared;  // Square of distance between object and left bullet
            rSquaredRight = (dxRight * dxRight) + dySquared;  // Square of distance between object and right bullet
            // Is square of distance between objects <= sum of object radii squared?  (Yes == hit).
            return ((rSquaredLeft <= (this.rSquared + object2.rSquared)) ||
                    (rSquaredRight <= (this.rSquared + object2.rSquared)));
        };
        bullet.draw = function() {
            gv_context.drawImage(this.image, this.x - gv_bulletSeparation - this.w2, this.y - this.h2);
            gv_context.drawImage(this.image, this.x + gv_bulletSeparation - this.w2, this.y - this.h2);
        };

        gv_bullets.push(bullet);
    }
}

function initEnemies() {
    var enemy;
    var ii;

    gv_enemies = [];

    for (ii = 0; ii < gv_numEnemies; ii++) {
        enemy = new Actor(gv_images.enemy, "down");
        enemy.speed = gv_enemySpeed;
        enemy.friction = 1.0;
        enemy.isSpawned = false;
        enemy.isAlive = false;
        enemy.getRandomStart = function() {
            this.x = utils.getRandomInt(0, gv_canvas.width);
            this.y = - this.h2 - this.h2;
        };
        enemy.edgeDetect = function() {
            if (this.y >= gv_canvas.height) {
                this.getRandomStart();
            }
        };

        gv_enemies.push(enemy);
    }
}

function startGame() {
    gv_isGameEnding = false;
    gv_isGameInProgress = true;

    initShip();
    initBullets();
    initEnemies();
    spawnEnemy();
    log("");  // Clear the log text

    (function drawFrame() {
        gv_requestAnimationFrame = window.requestAnimationFrame(drawFrame, gv_canvas);
        updateGame();
    }());
}

function addListeners() {
    var startButton = document.getElementById("start_button");
    var soundOnRadioButton = document.getElementById("sound_on");
    var soundOffRadioButton = document.getElementById("sound_off");
    var difficultySelector = document.getElementById("difficulty");

    window.addEventListener("keydown", onKeyDown, false);
    window.addEventListener("keyup", onKeyUp, false);

    startButton.onclick = function() {
        gv_isSoundOn = soundOnRadioButton.checked;
        setDifficulty(difficultySelector.value);
        difficultySelector.disabled = true;
        this.style.visibility = "hidden";
        startGame();
    };
    difficultySelector.onchange = function() {
        setDifficulty(this.value);
    };
    soundOnRadioButton.onclick = function() {
        gv_isSoundOn = this.checked;
    };
    soundOffRadioButton.onclick = function() {
        gv_isSoundOn = (! this.checked);
    };

    soundOnRadioButton.checked = gv_isSoundOn;
    startButton.disabled = false;
}

/**
 * Initialize the game.
 */
function initGame() {
    gv_ship = new Actor(gv_images.ship, "right");

    gv_ship.x = gv_canvas.width / 2;
    gv_ship.y = gv_canvas.height - 80;

    gv_ship.edgeDetect = function() {
        if (this.x - this.w2 <= 0) {
            this.x = this.w2;
            this.direction = "right";
            this.computeAngle();
        } else if (this.x + this.w2 >= gv_canvas.width) {
            this.x = gv_canvas.width - this.w2;
            this.direction = "left";
            this.computeAngle();
        }
    };

    gv_ship.updateSpeed = function() {
        if (this.isMovingRight) {
            if (this.direction == "right") {
                this.speed += this.deltaAccel;
            } else if (this.direction == "left") {
                this.speed -= this.deltaAccel;
            }
        } else if (this.isMovingLeft) {
            if (this.direction == "right") {
                this.speed -= this.deltaAccel;
            } else if (this.direction == "left") {
                this.speed += this.deltaAccel;
            }
        }
    };

    addListeners();

    doUpdateGame();  // Show ship in starting location.
}

/**
 * This function is called once for each image loaded.
 */
function onImageLoaded() {
    gv_numImagesLoaded++;

    // start game only after we know all images have been loaded.
    if (gv_numImagesLoaded == gv_numImagesRequested) {
        initGame();
    }
}

/**
 * Load an image given the name of the PNG image file.
 * @param imageName The name of the image file without its file suffix (assumes a ".png" file suffix).
 */
function loadImage(imageName) {
    var image = new Image();
    gv_images[imageName] = image;

    image.onload = onImageLoaded;
    image.src = "images/" + imageName + ".png";
}

// Load the images.  This game is started after the last image completes loading.
loadImage("bullet");
loadImage("explosion");
loadImage("enemy");
loadImage("ship");
